{-# OPTIONS --sized-types #-}
{- 2010-02-02 Cody Roux communicated and observation of Frederic
   Blanqui that the "non-linear" size-assignment for constructors (see
   M below) does not allow to express the precise sizes in a deep
   match involving a limit ordinal (see L below).  From this I could
   construct a non-looping term in MiniAgda -}

-- Agda is not affected by this bug

module Cody where

open import Size

data Nat : Set where
  zero : Nat
  succ : Nat -> Nat


data O : (_ : Size) -> Set where
  Z : {i : Size} -> O (↑ i)
  S : {i : Size} -> O i -> O (↑ i)
  L : {i : Size} -> (Nat -> O i) -> O (↑ i)
  M : {i : Size} -> O i -> O i -> O (↑ i)
 

emb : Nat -> O ∞
emb zero = Z
emb (succ n) = S (emb n)

-- refuted by Agda
caseO : {i : Size} -> O (↑ i) -> {C : Size -> Set} ->
            C i ->
            (O i -> C i) ->
            ((Nat -> O i) -> C i) ->
            (O i -> O i -> C i) ->
            C i
caseO  Z      z s l m = z
caseO (S x)   z s l m = s x
caseO (L g)   z s l m = l g
caseO (M a b) z s l m = m a b

pre : {i : Size} -> (Nat -> O (↑ ↑ i)) -> Nat -> O (↑ i)
pre {i} f n with (f (succ n))
-- ... | (Z .{↑ i}) = Z {i}
... | Z = Z
... | (S x) = x
... | (L g) = g n
... | (M a b) = a

{- the following pattern match is the problem: (L .(↑↑ i) f)

   The correct pattern is (L .i f).  By subtyping these two terms are
   both of the required type O (↑↑↑ i), but the least size must
   be chosen to be on the safe side.

   Thus, the implementation of dot patterns has to be changed to account
   for subtyping.

-}

deep : {i : Size} -> O i -> Nat -> Nat
deep (M (L f) (S (S (S x)))) n = 
  deep (M (L (pre f)) (S (f n))) (succ (succ (succ n)))
deep i x n = n   

four : Nat
four = succ (succ (succ (succ zero)))

loop : Nat 
loop = deep (M (L emb) (emb four)) four